%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% MCDCTrain.m
%
% Usage (execute newly)
%
% [IDX, SKP, OKP, FV, GV, XE, YE, loglik] = MCDCTrain(algorithm, ...
%                                                     xGrids, ...
%                                                     stateKernelGens, obsKernelGens, ...
%                                                     stateMeanFuncs, obsMeanFuncs, ...
%                                                     gridDimForGramMatrix, ...
%                                                     splineHandle, ...
%                                                     x0, u, z, N, J, K, ...
%                                                     aspect)
%
% input:
%   algorithm:  Algorithm object （e.g., PSMC, PMCMC）
%   xGrids: grid data for state variable (cell matrix for each dimention)
%   stateKernelGens: cell matrix of objects to generate kernel function for each dimension of state variable (cell matrix of 'KernelGenerator')
%   obsKernelGens: cell matrix of objects to generate kernel function for each dimension of observation variable (cell matrix of 'KernelGenerator')
%   stateMeanFuncs: cell matrix of mean function of state variable for each dimention
%   obsMeanFuncs: cell matrix of mean function of observation variable for each dimention
%   gridDimForGramMatrix: dimension to generate kernel covariance matrix (used only for PMCMC2)
%   splineHandle: function handler to generate spline interpolation function
%   x0: initial value for state variable
%   xaux: auxiliary variable for state variable
%   u: control input for state variable
%   z: time series data of observation values
%   N: number of particles for particle filter
%   J: iteratrion for MCMC step
%   K: time difference between state and observation
%   aspect: Aspect object
%
% output:
%   SKP: parameters for kernel function of state variable through the iterations
%   OKP: parameters for kernel function of observation variable through the iterations
%   FV: state values on each grid through the iteration
%   GV: observation values on each grid through the iteration
%   XE: transition of state variable through the iteration
%   YE: transition of observation variable through the iteration
%   loglik: transition of log-likehood through the iteration
%
%
% Usage (execute consecutively)
%
% [IDX, SKP, OKP, FV, GV, XE, YE, loglik] = MCDCTrain(matfile, aspect, Name, Value. ...)
%
% input:
%   matfile: .mat file for intermidiate result of previous execution
%   aspect: 'Aspect' object
%   Name, Value, ...: pairs of arguments if changing parameters of previous execution is required
%     'N': number of particle for 'Particle Filter' module
%
% output:
%   same as 'newly execution'
%
function [ IDX, SKP, OKP, FV, GV, XE, YE, loglik ] = MCDCTrain(varargin)
  if isa(varargin{1}, 'Algorithm')
    algorithm = varargin{1};
    in = MCDCInput(varargin{2:15});
    out = MCDCOutput(in.J);
    aspect = varargin{16};
  elseif ischar(varargin{1})
    [algorithm, in, out] = MCDCMatFile.Load(varargin{1});
    aspect = varargin{2};
    overrideInput(in, varargin{:});
  end

  algorithm.Execute(in, out, aspect);
  IDX = out.IDX;
  SKP = out.SKP;
  OKP = out.OKP;
  FV = out.FV;
  GV = out.GV;
  XE = out.XE;
  YE = out.YE;
  loglik = out.loglik;

end

function overrideInput(in, varargin)
  n = 3;
  while (nargin >= (n + 1))
    key = varargin{n};
    if (strcmp(key, 'N'))
      in.N = varargin{n + 1};
    end
    n = n + 2;
  end
end
