%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% PSMC.m
%
% input:
%   xGrids: grid data for state variable (cell matrix for each dimention)
%   initialStateKernelGens: cell matrix of objects to generate initial kernel function for each dimension of state variable (cell matrix of 'KernelGenerator')
%   initialObsKernelGens: cell matrix of objects to generate initial kernel function for each dimension of observation variable (cell matrix of 'KernelGenerator')
%   initialStateMeanFuncs: cell matrix of initial mean function of state variable for each dimention
%   initialObsMeanFuncs: cell matrix of initial mean function of observation variable for each dimention
%   gridDimForGramMatrix: dimension to generate kernel covariance matrix (used only for PMCMC2)
%   splineHandle: function handler to generate spline interpolation function
%   x0: initial value for state variable
%   xaux: auxiliary variable for state variable
%   u: control input for state variable
%   z: time series data of observation values
%   N: number of particles for particle filter
%   J: iteratrion for MCMC step
%   K: time difference between state and observation
%   aspect: Aspect object
%
% output:
%   SKP: parameters for kernel function of state variable through the iterations
%   OKP: parameters for kernel function of observation variable through the iterations
%   FV: state values on each grid through the iteration
%   GV: observation values on each grid through the iteration
%   XE: transition of state variable through the iteration
%   YE: transition of observation variable through the iteration
%   loglik: transition of log-likehoood through the iteration

classdef PSMC < Algorithm

  methods (Access = public)

    function this = PSMC()
      this = this@Algorithm(@RBFKernelGeneratorStrategyChoice1, @MCDCStrategyChoice1_PSMC);
    end

    function Execute(this, in, out, aspect)
      xdim = in.XDim();
      ydim = in.YDim();
      T = in.T();

      this.stateKGenStrategy.SetInitialKernelGens(in.stateKernelGens);
      this.obsKGenStrategy.SetInitialKernelGens(in.obsKernelGens);

      gridData = GridData(in.xGrids);
      this.mcdcStrategy.SetGridData(gridData);

      % Start
      aspect.Start(this, in, out);
      cleanupObj = onCleanup(@() aspect.End());

      % Grid Check
      CheckGridTransformation(VectorValuedFunction(in.stateMeanFuncs), gridData, aspect);

      %%%
      %%% MCMC Iteration
      %%%

      for j = (out.j + 1):in.J

        aspect.StartIteration();

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 1: Sample kernel parameters
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        stateKernelGens = this.stateKGenStrategy.GetGenerator(j, in, out);
        obsKernelGens = this.obsKGenStrategy.GetGenerator(j, in, out);

        for i = 1:xdim
          stateKernels{i} = stateKernelGens{i}.Generate();
          aspect.Log('  StateKernel[%d]: %s', i, stateKernels{i}.ToString());
        end
        for i = 1:ydim
          obsKernels{i} = obsKernelGens{i}.Generate();
          aspect.Log('  ObsKernel[%d]: %s', i, obsKernels{i}.ToString());
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 3: Construct gram matrix on grid points
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        aspect.Log('  Creating GramMatrix using all dims...');

        gramMatrixGrid = gridData.GetFlatMatrix();

        for i = 1:xdim
          GramMatState{i} = stateKernels{i}.CreateGramMatrix(gramMatrixGrid);
          aspect.Log('    StateKernel[%d] Done', i);
        end

        for i = 1:ydim
          GramMatObs{i} = obsKernels{i}.CreateGramMatrix(gramMatrixGrid);
          aspect.Log('    ObsKernel[%d] Done', i);
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 4: Draw realization of the GP surface
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        aspect.Log('  Drawing GP surface...');

        StateEqnVals = this.DrawGPSurface( ...
            gridData, ...
            in.gridDimForGramMatrix, ...
            this.mcdcStrategy.StateMean(j, in, out), ...
            this.mcdcStrategy.StateSigma(j, GramMatState));
        ObsEqnVals = this.DrawGPSurface( ...
            gridData, ...
            in.gridDimForGramMatrix, ...
            this.mcdcStrategy.ObsMean(j, in, out), ...
            this.mcdcStrategy.ObsSigma(j, GramMatObs));

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 5: Construct a B3 spline approximation
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        FnState = VectorValuedFunction(ModelFunctions2(in.splineHandle, gridData, StateEqnVals));
        FnObs = VectorValuedFunction(ModelFunctions2(in.splineHandle, gridData, ObsEqnVals));

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 6: Estimate using particle filter
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        aspect.Log('  Estimating using particle filter... (# of particles = %d)', in.N);

        q = 1.0;
        r = 0.1;

        particleFilter = this.mcdcStrategy.GetParticleFilter();
        pf = particleFilter(in.x0, in.xaux', in.u', in.z', FnState, FnObs, q, r, in.N, in.K);

        acceptanceLogProbability = this.mcdcStrategy.AcceptanceLogProbability( ...
            j, in, out, pf.Loglik, StateEqnVals, ObsEqnVals, GramMatState, GramMatObs);
        aspect.Log('  Acceptance log probability = %f', acceptanceLogProbability);

        if rand(1) < exp(acceptanceLogProbability)
          out.AddSKP(cell2mat(arrayfun(@(i) stateKernels{i}.Properties(), 1:xdim, 'UniformOutput', false))');
          out.AddOKP(cell2mat(arrayfun(@(i) obsKernels{i}.Properties(), 1:ydim, 'UniformOutput', false))');
          out.AddFV(StateEqnVals);
          out.AddGV(ObsEqnVals);
          ji = find(rand(1) < cumsum(pf.Weights(:, T - in.K)), 1, 'first');
          xe = reshape(pf.Particles(ji, :, :), size(pf.Particles, 2), size(pf.Particles, 3));
          out.AddXE(xe);
          if length(in.xaux) > 0
            out.AddYE(FnObs([xe in.xaux']));
          else
            out.AddYE(FnObs(xe));
          end
          out.AddLoglik(pf.Loglik);
          if (mod(j, 100) == 0)
              out.AddWeight(pf.Weights);
          end
          out.AddFnState(FnState, j);
          out.AddFnObs(FnObs, j);

          this.mcdcStrategy.SaveGramMatrices(GramMatState, GramMatObs);

          if j == 1
            aspect.Log('  log-likelihood = %f', pf.Loglik);
          else
            aspect.Log('  log-likelihood = %f, accepted', pf.Loglik);
          end
        else
          aspect.Log('  log-likelihood = %f, rejected', pf.Loglik);
        end

        out.SetIDXAsLast(j);
        out.j = j;
        aspect.EndIteration();

      end

    end

    function result = DrawGPSurface(this, gridData, pivot, means, gramMatrices)
      dim = size(means, 1);
      result = zeros([dim gridData.CountPerDim]);
      indices = gridData.GetAllIndices();
      for d = 1:dim
        for i = 1:size(indices, 1)
          result(d, indices(i, :)) = mvnrnd(means(d, indices(i, :)), gramMatrices{d});
        end
      end
    end

    function result = CalcLogProbability(this, values, gridData, pivot, means, gramMatrices)
      dim = size(means, 1);
      result = 0;
      indices = gridData.GetAllIndices();
      for d = 1:dim
        for i = 1:size(indices, 1)
          x = values(d, indices(i, :));
          mu = means(d, indices(i, :));
          sigma = gramMatrices{d};
          eigen = eig(sigma);
          if min(eigen) < 1.0e-6
            sigma = sigma + (min(eigen) + 0.1).* eye(size(sigma));
          end
          result = result + LogMvnPdf(x, mu, sigma);
        end
      end
    end

  end

end
